#include <avr/io.h>
#include <avr/interrupt.h>
#define F_CPU 16000000
#include <util/delay.h>
#include <stdlib.h>

#define LATCH0 PA1
#define DATA0 PA0
#define CLK0 PA2

#define LATCH1 PD1
#define DATA1 PD0
#define CLK1 PD2

#define START0_R 4
#define START0_C 3
#define START1_R 4
#define START1_C 4

#define LEFT 1
#define RIGHT 2
#define UP 3
#define DOWN 4

#define FOOD 1
#define SNAKE 2

#define FOOD_TIME 10
#define FOOD_LIFE 14

typedef struct position {
	unsigned char row, column;
} position_t;

volatile char direction;
volatile char next_direction;
volatile char button_pressed;

position_t snake[128];
unsigned char snake_size;

unsigned char map[8][16];
position_t food_p[50];
unsigned char food_locations = 50;
unsigned char food_life, food_exists, food_count, food_eaten, food_time;

uint32_t delay_tick, DELAY;

void clear_display() {

	char i;

	PORTA &= ~(1 << LATCH0);
	PORTA &= ~(1 << CLK0);
	for (i = 0; i < 8; ++i) {
		PORTA &= ~(1 << DATA0);
		
		PORTA |= (1 << CLK0);
		PORTA &= ~(1 << CLK0);
	}
	
	for (i = 0; i < 8; ++i) {
		PORTA &= ~(1 << DATA0);
		
		PORTA |= (1 << CLK0);
		PORTA &= ~(1 << CLK0);
	}
	
	PORTA |= (1 << LATCH0);
	
	PORTD &= ~(1 << LATCH1);
	PORTD &= ~(1 << CLK1);
	for (i = 0; i < 8; ++i) {
		PORTD &= ~(1 << DATA1);
		
		PORTD |= (1 << CLK1);
		PORTD &= ~(1 << CLK1);
	}
	
	for (i = 0; i < 8; ++i) {
		PORTD &= ~(1 << DATA1);
		
		PORTD |= (1 << CLK1);
		PORTD &= ~(1 << CLK1);
	}
	
	PORTD |= (1 << LATCH1);
}

void refresh() {
	int i, j;
	char line;
	
	for (i = 0; i < 8; ++i) {
		line = 1 << i;
		unsigned char *v;
	
		v = &map[i][0];
		PORTA &= ~(1 << LATCH0);
		//PORTA &= ~(1 << CLK0);
		for (j = 7; j >= 0; --j) {
			if (v[j] > 0) {
				PORTA |= (1 << DATA0);
			}
			else {
				PORTA &= ~(1 << DATA0);
			}
			
			PORTA |= (1 << CLK0);
			PORTA &= ~(1 << CLK0);
		}
		
		for (j = 0; j < 8; ++j) {
			if (((line >> j) & 1) == 1) {
				PORTA |= (1 << DATA0);
			}
			else {
				PORTA &= ~(1 << DATA0);
			}
			
			PORTA |= (1 << CLK0);
			PORTA &= ~(1 << CLK0);
		}
		
		PORTA |= (1 << LATCH0);
		
		v = &map[i][8];
		PORTD &= ~(1 << LATCH1);
		for (j = 7; j >= 0; --j) {
			if (v[j] > 0) {
				PORTD |= (1 << DATA1);
			}
			else {
				PORTD &= ~(1 << DATA1);
			}
			
			PORTD |= (1 << CLK1);
			PORTD &= ~(1 << CLK1);
		}
		
		for (j = 0; j < 8; ++j) {
			if (((line >> j) & 1) == 1) {
				PORTD |= (1 << DATA1);
			}
			else {
				PORTD &= ~(1 << DATA1);
			}
			
			PORTD |= (1 << CLK1);
			PORTD &= ~(1 << CLK1);
		}
		
		
		PORTD |= (1 << LATCH1);
	}
}

void init_io() {
	DDRA |= (1 << LATCH0) | (1 << DATA0) | (1 << CLK0);
	DDRD |= (1 << LATCH1) | (1 << DATA1) | (1 << CLK1);
	
	DDRB &= ~( (1 << PB0) | (1 << PB1) | (1 << PB2) | (1 << PB3));
    PORTB |= ( (1 << PB0) | (1 << PB1) | (1 << PB2) | (1 << PB3));
	
	PCICR |= (1 << PCIE1);
    PCMSK1 |= (1 << PCINT8) | (1 << PCINT9) | (1 << PCINT10) | (1 << PCINT11);

	sei();
}

void init_game() {
	unsigned char i, j;
	
	clear_display();
	
	for (i = 0; i < 8; ++i) {
		for (j = 0; j < 16; ++j) {
			map[i][j] = 0;
		}
	}
	
	DELAY = 800;
	
	snake[0].row = START0_R;
	snake[0].column = START0_C;
	snake[1].column = START1_C;
	snake[1].row = START1_R;
	snake_size = 2;
	
	map[START0_R][START0_C] = 1;
	map[START1_R][START1_C] = 1;
	direction = next_direction = RIGHT;
	
	food_exists = 0;
	food_count = 0;
	food_eaten = 0;
	food_time = FOOD_TIME;
	
	for (i = 0; i < food_locations; ++i) {
		food_p[i].column = rand() % 16;
		food_p[i].row = rand() % 8;
	}
}


ISR(PCINT1_vect) {

	button_pressed = 1;

	if ((PINB & (1 << PB0)) == 0) {
		if (direction == LEFT || direction == RIGHT) {
			next_direction = UP;
		}
	}
	
	
	if ((PINB & (1 << PB1)) == 0) {
		if (direction == UP || direction == DOWN) {
			next_direction = RIGHT;
		}
	}
	
	if ((PINB & (1 << PB2)) == 0) {
		if (direction == LEFT || direction == RIGHT) {
			next_direction = DOWN;
		}
	}
	
	if ((PINB & (1 << PB3)) == 0) {
		if (direction == UP || direction == DOWN) {
			next_direction = LEFT;
		}
	}
}

void spawn_food() {

	if (food_count == food_locations) {
		food_count = 0;
	}
	
	if (map[food_p[food_count].row][food_p[food_count].column] != 0) {
		++food_count;
		return;
	}

	
	map[food_p[food_count].row][food_p[food_count].column] = FOOD;
	++food_count;
	food_exists = 1;
	food_life = FOOD_LIFE;
}

void delete_food() {
	map[food_p[food_count-1].row][food_p[food_count-1].column] = 0;
	food_exists = 0;
	food_time = FOOD_TIME;
}

void wait_button_press() {
	button_pressed = 0;
	while (button_pressed == 0) {
		refresh();
	}
	
	return;
}

void compensate_delay() {
	DELAY -= 4;
}

void play_game() {
	unsigned char i, limit;
	position_t snake_tail;
	position_t snake_head;

	while(1) {
		delay_tick = DELAY;
		while(--delay_tick) {
			refresh();
		}

		snake_tail = snake[0];
		snake_head = snake[snake_size - 1];
		limit = snake_size - 1;

		if (direction == RIGHT) {
			if (snake[limit].column == 15) {
				snake[limit].column = 0;
			}
			else {
				snake[limit].column += 1;
			}
		}
		else if (direction == LEFT) {
			if (snake[limit].column == 0) {
				snake[limit].column = 15;
			}
			else {
				snake[limit].column -= 1;
			}
		}
		else if (direction == UP) {
			if (snake[limit].row == 0) {
				snake[limit].row = 7;
			}
			else {
				snake[limit].row -= 1;
			}
		}
		else if (direction == DOWN) {
			if (snake[limit].row == 7) {
				snake[limit].row = 0;
			}
			else {
			snake[limit].row += 1;
			}
		}
		
		map[snake_tail.row][snake_tail.column] = 0;
		
		if ((i = map[snake[limit].row][snake[limit].column]) > 0) {	//moare (verifica pentru mancare!!!)
			if (i == FOOD) {
				food_eaten = 1;
			}
			else {
				return;
			}
		}
		
		if (food_eaten == 1) {
			snake[snake_size] = snake[snake_size - 1];
			snake[snake_size - 1] = snake_head;
			
			delete_food();
			
			map[snake_tail.row][snake_tail.column] = SNAKE;
			map[snake[snake_size].row][snake[snake_size].column] = SNAKE;
			++snake_size;
			compensate_delay();
			
			food_eaten = 0;
		}
		else {
			for (i = 0; i < limit - 1; ++i) {
				snake[i] = snake[i+1];
			}
			
			snake[i] = snake_head;
			map[snake[limit].row][snake[limit].column] = SNAKE;
			
			if (food_time > 0) {
				--food_time;
			}
			else if (food_exists == 0) {
				spawn_food();
			}
			else {
				--food_life;
				if (food_life == 0) {
					delete_food();
				}
			}
		}
		direction = next_direction;
	}
}

int main(void) {
	
	init_io();
	
	while(1) {
		init_game();
		wait_button_press();
		play_game();
	}
	return 0;
}

